from tfcat.base import Base
from astropy.units import Unit, Quantity


class Field(Base):

    def __init__(self, info=None, datatype=None, ucd=None, unit=None, **extra):
        super(Field).__init__(**extra)
        self.info = info
        self.datatype = datatype
        self.ucd = ucd
        self.unit = unit



from tfcat.base import Base


class Observation(Base):
    """
    Represents a TF observation
    """

    def __init__(self, # id=None,
                 geometry=None, properties=None, **extra):
        """
        Initialises a Feature object with the given parameters.

        :param id: Feature identifier, such as a sequential number.
        :type id: str, int
        :param geometry: Geometry corresponding to the feature.
        :param properties: Dict containing properties of the feature.
        :type properties: dict
        :return: Observation object
        :rtype: Observation
        """
        super(Observation, self).__init__(**extra)
        # if id is not None:
        #    self["id"] = id
        self["geometry"] = (self.to_instance(geometry, strict=True)
                            if geometry else None)
        self["properties"] = properties or {}

    def errors(self):
        geo = self.get('geometry')
        return geo.errors() if geo else None


class ObservationCollection(Base):
    """
    Represents a FeatureCollection, a set of multiple Feature objects.
    """

    def __init__(self, observations, **extra):
        """
        Initialises a FeatureCollection object from the

        :param observations: List of features to constitute the FeatureCollection.
        :type observations: list
        :return: ObservationCollection object
        :rtype: ObservationCollection
        """
        super(ObservationCollection, self).__init__(**extra)
        self["observations"] = observations

    def errors(self):
        return self.check_list_errors(lambda x: x.errors(), self.features)

    def __getitem__(self, key):
        try:
            return self.get("observations", ())[key]
        except (KeyError, TypeError, IndexError):
            return super(Base, self).__getitem__(key)




from tfcat.base import Base
from astropy.time import Time
from astropy.units import Unit


class CRS(Base):

    def __init__(self, crs=None, type="local", properties=None, **extra):
        super(CRS).__init__(**extra)
        if crs is not None:
            self['type'] = crs['type']
            self['properties'] = crs['properties']
        if properties is not None:
            self['type'] = type
            self['properties'] = properties

    @classmethod
    def configure(
            cls,
            spectral_coords_id: str,
            ref_position_id: str,
            time_coords=None,
            time_coords_id=None,
            crs_name=None
    ):
        if time_coords is None and time_coords_id is None:
            raise ValueError("TFCat CRS input parameters error: Either 'time_coords' or 'time_coords_id' must be set.")

        crs = {
            'type': "local",
            'properties': {
                'spectral_coords': SPECTRAL_COORDS[spectral_coords_id],
                'ref_position_id': ref_position_id,
            }
        }
        if crs_name is not None:
            crs["properties"]["name"] = crs_name

        if time_coords_id in time_mapping.keys():
            crs["properties"]["time_coords_id"] = time_coords_id
        else:
            crs["properties"]["time_coords"] = time_coords
        return CRS(crs=crs)

    @property
    def time_converter(self):
        if 'time_coords_id' in self.properties.keys():
            return time_mapping[self.properties['time_coords_id']]
        else:
            return self._custom_time_converter

    @property
    def spectral_converter(self):
        return lambda x: x * Unit(self.properties['spectral_coords']['unit'])

    @property
    def converter(self):
        return lambda x: (self.time_converter(x[0]), self.spectral_converter(x[1]))

    @property
    def _custom_time_converter(self):
        time_coords = self.properties['time_coords']
        time_origin = time_coords['time_origin']
        if 'T' in time_origin:
            time_format = 'isot'
        else:
            time_format = 'iso'
        return lambda t: (
            Time(time_origin, format=time_format, scale=time_coords['time_scale'].lower()) +
            t * Unit(time_coords['unit'])
        )


time_mapping = {
    'unix': lambda t: Time(t, format="unix"),
    'jd': lambda t: Time(t, format='jd'),
    'mjd': lambda t: Time(t, format='mjd'),
    'mjd_cnes': lambda t: Time("1950-01-01T00:00:00.000Z", format='iso') + t * Unit('d'),
    'mjd_nasa': lambda t: Time("1968-05-24T00:00:00.000Z", format='iso') + t * Unit('d'),
    'iso': lambda t: Time(t, format='iso'),
    'cdf_tt2000': lambda t: Time('2000-01-01 00:00:00.000Z', format='iso') + t * Unit('ns')
}

TIME_COORDS = {
    'unix': {
        "name": "Timestamp (Unix Time)",
        "unit": "s",
        "time_origin": "1970-01-01T00:00:00.000Z",
        "time_scale": "UTC"
    },
    'jd': {
        "name": "Julian Day",
        "unit": "d",
        "time_origin": "U",
        "time_scale": "UTC"
    },
    'mjd': {
        "name": "Modified Julian Day",
        "unit": "d",
        "time_origin": "1858-11-17T00:00:00.000Z",
        "time_scale": "UTC"
    },
    'mjd_cnes': {
        "name": "Modified Julian Day (CNES definition)",
        "unit": "d",
        "time_origin": "1950-01-01T00:00:00.000Z",
        "time_scale": "UTC"
    },
    'mjd_nasa': {
        "name": "Modified Julian Day (NASA definition)",
        "unit": "d",
        "time_origin": "1968-05-24T00:00:00.000Z",
        "time_scale": "UTC"
    },
    'cdf_tt2000': {
        "name": "CDF Epoch TT2000",
        "unit": "ns",
        "time_origin": "2000-01-01 00:00:00.000Z",
        "time_scale": "TT"
    }
}

SPECTRAL_COORDS = {
    'Hz': {
        "type": "frequency",
        "unit": "Hz"
    },
    'kHz': {
        "type": "frequency",
        "unit": "kHz"
    },
    'MHz': {
        "type": "frequency",
        "unit": "MHz"
    },
    'm': {
        "type": "wavelength",
        "unit": "m"
    },
    'cm': {
        "type": "wavelength",
        "unit": "cm"
    },
    'mm': {
        "type": "wavelength",
        "unit": "mm"
    },
    'cm-1': {
        "type": "wavenumber",
        "unit": "cm**-1"
    },
    'eV': {
        "type": "energy",
        "unit": "eV"
    },
}

REF_POSITION = [
    'GEOCENTER',
]

DefaultCRS = CRS({
    "type": "local",
    "properties": {
        "time_coords_id": 'unix',
        "spectral_coords": SPECTRAL_COORDS['Hz'],
        "ref_position_id": 'GEOCENTER',
    }
})



import jsonschema
from json import loads, load
import urllib.request
import ssl

_context = ssl._create_unverified_context()
JSONSCHEMA_URI = "https://voparis-ns.obspm.fr/maser/tfcat/v1.0/schema#"
TOPCAT_TFCAT_URL = "http://andromeda.star.bristol.ac.uk/releases/topcat/pre/topcat-full_tfcat.jar"


def tfcat_schema(schema_uri=JSONSCHEMA_URI):
    with urllib.request.urlopen(schema_uri, context=_context) as f:
        return loads(f.read().decode())


def validate(instance, schema_uri=JSONSCHEMA_URI):
    """Validate TFCat object

    :param instance: TFCat object
    :param schema_uri: URI of TFCat JSON schema
    """
    jsonschema.validate(instance, schema=tfcat_schema(schema_uri))


def validate_file(file_name, schema_uri=JSONSCHEMA_URI):
    """Validate TFCat data from file.

    :param file_name: TFCat data file name
    :type file_name: Path or str
    :param schema_uri: URI of TFCat JSON schema
    :type schema_uri: str
    """
    with open(file_name, 'r') as f:
        validate(load(f), schema_uri)



from tfcat.geometry import Point, LineString, Polygon
from tfcat.geometry import MultiLineString, MultiPoint, MultiPolygon
from tfcat.geometry import GeometryCollection
from tfcat.feature import Feature, FeatureCollection
from tfcat.base import Base
from tfcat.crs import CRS
#from tfcat.tfcat import TFCat

__all__ = ([Point, LineString, Polygon] +
           [MultiLineString, MultiPoint, MultiPolygon] +
           [GeometryCollection] +
           [Feature, FeatureCollection] +
           [Base] +
#           [TFCat] +
           [CRS])



import json
import tfcat.factory
import tfcat.base

"""
This file is adapted from: https://github.com/jazzband/geojson/blob/master/geojson/codec.py
"""


class TFCatEncoder(json.JSONEncoder):

    def default(self, obj):
        return tfcat.factory.Base.to_instance(obj) # NOQA


# Wrap the functions from json, providing encoder, decoders, and
# object creation hooks.
# Here the defaults are set to only permit valid JSON as per RFC 4267

def _enforce_strict_numbers(obj):
    raise ValueError("Number %r is not JSON compliant" % obj)


def dump(obj, fp, cls=TFCatEncoder, allow_nan=False, **kwargs):
    return json.dump(obj, fp, cls=cls, allow_nan=allow_nan, **kwargs)


def dumps(obj, cls=TFCatEncoder, allow_nan=False, **kwargs):
    return json.dumps(obj, cls=cls, allow_nan=allow_nan, **kwargs)


def load(fp,
         cls=json.JSONDecoder,
         parse_constant=_enforce_strict_numbers,
         object_hook=tfcat.base.Base.to_instance,
         **kwargs):
    return json.load(fp,
                     cls=cls, object_hook=object_hook,
                     parse_constant=parse_constant,
                     **kwargs)


def loads(s,
          cls=json.JSONDecoder,
          parse_constant=_enforce_strict_numbers,
          object_hook=tfcat.base.Base.to_instance,
          **kwargs):
    return json.loads(s,
                      cls=cls, object_hook=object_hook,
                      parse_constant=parse_constant,
                      **kwargs)





"""
This class is adapted from: https://github.com/jazzband/geojson/blob/master/geojson/feature.py
"""

from tfcat.base import Base
from tfcat.crs import DefaultCRS, CRS
from tfcat.validate import JSONSCHEMA_URI
from astropy.table import Table, Column
from tfcat.utils import split_coords
from astropy.time import Time
from shapely.geometry import shape
from matplotlib import pyplot as plt


class Feature(Base):
    """
    Represents a TF feature
    """ 

    def __init__(self, id=None,
                 geometry=None, properties=None, **extra):
        """
        Initialises a Feature object with the given parameters.

        :param id: Feature identifier, such as a sequential number.
        :type id: str, int
        :param geometry: Geometry corresponding to the feature.
        :param properties: Dict containing properties of the feature.
        :type properties: dict
        :return: Feature object
        :rtype: Feature
        """
        super(Feature, self).__init__(**extra)
        self["id"] = id
        self["geometry"] = (self.to_instance(geometry, strict=True)
                            if geometry else None)
        self["properties"] = properties or {}

    def errors(self):
        geo = self.get('geometry')
        return geo.errors() if geo else None

    @property
    def tmin(self):
        return self.bbox[0]
        #return self.geometry.tmin #numpy.min([item[0] for item in self.geometry.coordinates]) if self.geometry else None

    @property
    def tmax(self):
        return self.bbox[2]
        #return self.geometry.tmax # numpy.max([item[0] for item in self.geometry.coordinates]) if self.geometry else None

    @property
    def fmin(self):
        return self.bbox[1]
        #return self.geometry.fmin # numpy.min([item[1] for item in self.geometry.coordinates]) if self.geometry else None

    @property
    def fmax(self):
        return self.bbox[3]
        #return self.geometry.fmax # numpy.max([item[1] for item in self.geometry.coordinates]) if self.geometry else None

    @property
    def bbox(self):
        #return [self.tmin, self.fmin, self.tmax, self.fmax] if self.geometry else None
        return shape(self.geometry).bounds if self.geometry else None

    def __len__(self):
        return len(self.geometry.coordinates) if self.geometry else 0

    def _plot(self, crs=None):

        ftype = self['geometry']['type']
        coord = self['geometry']['coordinates']
        if ftype not in ['MultiLineString', 'MultiPolygon']:
            coord = [coord]

        plot_style = '+-'
        if ftype.endswith('Point'):
            plot_style = '+'

        for item in coord:
            itimes, ifreqs = split_coords(item, crs=crs)
            if crs is not None:
                plt.plot(itimes.datetime, ifreqs.value, plot_style)
            else:
                plt.plot(itimes, ifreqs, plot_style)

    def _plot_bbox(self, crs=None):
        bbox_times = [self.tmin, self.tmax, self.tmax, self.tmin, self.tmin]
        if crs is not None:
            bbox_times = crs.time_converter(bbox_times).datetime
        bbox_freqs = [self.fmin, self.fmin, self.fmax, self.fmax, self.fmin]
        plt.plot(bbox_times, bbox_freqs, '--', label='bbox')

    def plot(self, crs=None):

        self._plot(crs)
        self._plot_bbox(crs)

        if crs is not None:
            plt.xlabel(crs.properties['time_coords']['name'])
            plt.ylabel(f"{crs.properties['spectral_coords']['name']} ({crs.properties['spectral_coords']['unit']})")

        plt.title(f"({self['geometry']['type']})")

        plt.show()


class FeatureCollection(Base):
    """
    Represents a FeatureCollection, a set of multiple Feature objects.
    """

    def __init__(self, features=None, properties=None, fields=None, crs=None, schema=JSONSCHEMA_URI, **extra):
        """
        Initialises a FeatureCollection object from the

        :param features: List of features to constitute the FeatureCollection.
        :type features: list
        :return: FeatureCollection object
        :rtype: FeatureCollection
        """
        super(FeatureCollection, self).__init__(**extra)
        self["$schema"] = schema
        self["features"] = features
        self["fields"] = fields or {}
        self["properties"] = properties or {}
        self["crs"] = CRS(crs=crs) or DefaultCRS

    def errors(self):
        return self.check_list_errors(lambda x: x.errors(), self.features)

    def __getitem__(self, key):
        try:
            return self.get("features", ())[key]
        except (KeyError, TypeError, IndexError):
            return super(Base, self).__getitem__(key)

    def __len__(self):
        return len(self.features)

    def as_table(self):
        """
        produces a Astropy Table object containing the TFCat data

        :return:
        :rtype: Table
        """

        cols = [
            FeatureColumn(
                name='feature',
                data=[item.geometry for item in self.features],
                description='Feature',
                crs=self.crs
            ),
            Column(
                name='tmin',
                data=[self.crs.time_converter(item.bbox[0]).datetime for item in self.features],
                description='Feature Start Time'
            ),
            Column(
                name='tmax',
                data=[self.crs.time_converter(item.bbox[2]).datetime for item in self.features],
                description='Feature End Time'
            ),
            Column(
                name='fmin',
                data=[item.bbox[1] for item in self.features],
                unit=self.crs.properties['spectral_coords']['unit'],
                description='Feature lower spectral bound'
            ),
            Column(
                name='fmax',
                data=[item.bbox[3] for item in self.features],
                unit=self.crs.properties['spectral_coords']['unit'],
                description='Feature upper spectral bound'
            )
        ]

        for key in self.fields:
            cur_col = Column(
                name=key,
                dtype=self.fields[key]['datatype'],
                data=[item.properties[key] for item in self.features],
                unit=self.fields[key].get('unit', None),
                description=self.fields[key].get('info', None)
            )
            if 'values' in self.fields[key]:
                cur_col.meta = {'options': self.fields[key].get('values', None)}
            cols.append(cur_col)

        return Table(cols)


class FeatureColumn(Column):

    def __new__(cls, crs=None, **extra):
        self = super().__new__(cls, **extra)
        self._crs = crs or DefaultCRS
        return self

    @property
    def type(self):
        return Column([item['type'] for item in self])

    def coordinates(self, ifeature): #, time_format=None, spectral_unit=None):
        return split_coords(self[ifeature], self._crs)






from tfcat.codec import dump, dumps, load, loads, TFCatEncoder
from tfcat.utils import coords, split_coords #, map_coords
from tfcat.geometry import Point, LineString, Polygon
from tfcat.geometry import MultiLineString, MultiPoint, MultiPolygon
from tfcat.geometry import GeometryCollection
from tfcat.feature import Feature, FeatureCollection
from tfcat.observation import Observation, ObservationCollection
from tfcat.base import Base
from tfcat.crs import CRS
from tfcat._version import __version__, __version_info__
from tfcat.tfcat import TFCat
__all__ = (
        [dump, dumps, load, loads, TFCatEncoder] +
        [coords, split_coords] + # map_coords] +
        [Point, LineString, Polygon] +
        [MultiLineString, MultiPoint, MultiPolygon] +
        [GeometryCollection] +
        [Feature, FeatureCollection] +
        [Observation, ObservationCollection] +
        [Base] +
        [CRS] +
        [TFCat] +
        [__version__, __version_info__])



from tfcat.codec import load, loads
from tfcat.geometry import Polygon, MultiPolygon, Point, MultiPoint, LineString, MultiLineString
from tfcat.feature import Feature
import tfcat.utils
from matplotlib import pyplot as plt
from urllib.request import urlopen


class TFCat:

    def __init__(self, tfcat_data, file_name=None):
        self.file = file_name
        self.data = tfcat_data

    @property
    def data(self):
        return self._data

    @data.setter
    def data(self, tfcat_data):
        self._data = tfcat_data

    @property
    def has_observations(self):
        return 'observations' in self._data.keys()

    @property
    def _lookup(self):
        return list(item.properties['obs_id'] for item in self._data.features)

    @classmethod
    def from_file(cls, file_name):
        with open(file_name, 'r') as f:
            tfcat_data = load(f)
        return cls(tfcat_data, file_name=file_name)

    @classmethod
    def from_url(cls, url):
        tfcat_data = loads(urlopen(url).read())
        return cls(tfcat_data, file_name=url)

    @property
    def crs(self):
        return self._data.crs

    @property
    def properties(self):
        return self._data.properties

    @property
    def fields(self):
        return self._data.fields

    def __len__(self):
        return len(self._data.features)

    def observation(self, n):
        return self._data.observations[n]

    def feature(self, n):
        return self._data.features[n]

    @property
    def iter_features(self) -> Feature:
        """
        Generator looping on features
        :return: feature
        """
        n = 0
        while n < len(self):
            yield self.feature(n) #, self.observation(self._lookup[n]) if self.has_observations else None
            n += 1

    @property
    def iter_observations(self) -> Feature:
        """
        Generator looping on observations
        :return: observation
        """
        n = 0
        while n < len(self._data.observations):
            yield self.observation(n)
#            yield [self.feature(i) for i, x in enumerate(self._lookup) if x == n], self.observation(n)
            n += 1

    def iter_features_by_obs_id(self, obs_id: int) -> Feature:
        """
        Generator looping on features within a given obs_id
        :param obs_id: Observation obs_id
        :return: feature
        """
        for i, x in enumerate(self._lookup):
            if x == obs_id:
                yield self.feature(i)

    def _plot_observation(self, oid):

        obs = self.observation(oid)
        crs = self.crs

        bbox_times = [obs.tmin, obs.tmax, obs.tmax, obs.tmin, obs.tmin]
        bbox_times = crs.time_converter(bbox_times).datetime
        bbox_freqs = [obs.fmin, obs.fmin, obs.fmax, obs.fmax, obs.fmin]

        plt.plot(bbox_times, bbox_freqs, '--', label='bbox')

    _plot_style = {
        Point: '+',
        LineString: '-',
        Polygon: '-',
        MultiPoint: '+',
        MultiPolygon: '-',
        MultiLineString: '-',
    }

    def _plot_feature(self, fid):

        feature = self.feature(fid)
        crs = self.crs

        ftype = type(feature.geometry)
        coord = feature['geometry']['coordinates']
        if ftype not in [MultiLineString, MultiPolygon]:
            coord = [coord]
        for item in coord:
            itimes, ifreqs = tfcat.utils.split_coords(item, crs=crs)
            plt.plot(itimes.datetime, ifreqs.value, self._plot_style[ftype], label=f'Feature #{fid}')

    def plot_observation(self, obs_id):

        crs = self.crs

        self._plot_observation(obs_id)

        plt.xlabel(crs.properties['time_coords']['name'])
        plt.ylabel(f"{crs.properties['spectral_coords']['name']} ({crs.properties['spectral_coords']['unit']})")
        plt.title(f'Observation #{obs_id}')

        if obs_id in self._lookup:
            for fid in (i for i, x in enumerate(self._lookup) if x == obs_id):
                self._plot_feature(fid)

        plt.show()

    def plot_feature(self, fid):
        from tfcat.crs import TIME_COORDS

        crs = self.crs
        if crs.type == 'local':
            time_coords = crs.properties.get(
                'time_coords',
                TIME_COORDS[crs.properties['time_coords_id']]
            )
        else:
            raise NotImplemented()

        self._plot_feature(fid)

        plt.xlabel(time_coords['name'])
        plt.ylabel(f"{crs.properties['spectral_coords']['type']} ({crs.properties['spectral_coords']['unit']})")
        plt.title(f'Feature #{fid}')
        plt.show()

    def to_votable(self, file_xml='votable_tfcat.xml'):
        from astropy.io.votable.tree import Param
        from astropy.io.votable import from_table

        votable = from_table(self._data.as_table())
        table = votable.get_first_table()

        for name, value in self.properties.items():
            table.params.append(
                Param(votable, name=name, value=value, arraysize="*", datatype="char")
            )

        votable.to_xml(file_xml)

    def add_property(self, name, field_def, feature_values=None, observation_values=None):

        if not isinstance(field_def, dict):
            raise TypeError('field_def must be a dict')

        self._data.fields[name] = field_def

        if feature_values is not None:
            for feature, value in zip(self._data.features, feature_values):
                feature.properties[name] = value
        if observation_values is not None:
            for observation, value in zip(self._data.observations, observation_values):
                observation.properties[name] = value
